import csv
import json
import re
from typing import Dict, List, Union

from pb_analyzer.const import ResponseKeys


def count_hidden_true_in_dict(response: Dict) -> int:
    """
    Counts the number of occurrences of the key-value pair "'Hidden': True" in the given dictionary response,
    ignoring entities where the name starts with 'DateTableTemplate'.

    Args:
        response (Dict): The API response as a dictionary.

    Returns:
        int: The number of occurrences of "'Hidden': True" in the response, excluding entities named 'DateTableTemplate*'.
    """

    def recursive_count(data: Union[Dict, List], skip_entity: bool = False) -> int:
        count: int = 0
        if isinstance(data, dict):
            for key, value in data.items():
                if key == 'Name' and isinstance(value, str) and (
                        value.startswith('DateTableTemplate') or value.startswith('LocalDateTable')):
                    skip_entity = True
                if key == 'Hidden' and value is True and not skip_entity:
                    count += 1
                count += recursive_count(value, skip_entity)
        elif isinstance(data, list):
            for item in data:
                count += recursive_count(item, skip_entity)
        return count

    return recursive_count(response)


def _extract_tables_and_columns(response: dict):
    """
    Extracts the tables and columns from the conceptual schema response.

    Args:
        response (dict): The conceptual schema response as a dictionary.

    """
    columns = []
    for schema in response.get('schemas', [{}]):
        if schema.get('error'):
            break
        for entity in schema.get('schema', {}).get('Entities', []):
            table_name = entity.get('Name', 'UnknownTable')
            for prop in entity.get('Properties', []):
                column_name = prop.get('Name', 'UnknownColumn')
                columns.append({'table': table_name, 'column': column_name})

    return columns


def fetch_columns_and_tables(conceptual_schema: dict):
    """
    Extracts and filters the tables and columns from the conceptual schema response.

    Parameters:
    conceptual_schema (dict): The conceptual schema response as a dictionary.
    """
    extracted_cols_and_tables = _extract_tables_and_columns(conceptual_schema)
    filtered_cols_and_tables = [item for item in extracted_cols_and_tables if
                                "DateTableTemplate" not in item.get('table') and "LocalDateTable" not in item.get(
                                    'table')]
    return filtered_cols_and_tables


def get_classified_columns(report_columns: List[dict], json_string: dict, ) -> tuple[str, List[dict]]:
    """
    Filters out strings from the list that are not found in the JSON string.

    Args:
        report_columns (List[dict]): The list of strings to filter.
        json_string (dict): The JSON string to check against.

    Returns:
        tuple[str, str, List[dict]]: A tuple containing the unused columns and all columns.
    """
    json_str: str = json.dumps(json_string)
    normalized_json_string: str = json_str.replace('"', '').replace("'", '')
    unused_columns_and_tables = [column for column in report_columns if
                                 f'{column[ResponseKeys.TABLE]}.{column[ResponseKeys.COLUMN]}' not in normalized_json_string]
    tables = {}
    for column in report_columns:
        if column[ResponseKeys.TABLE] not in tables:
            tables[column[ResponseKeys.TABLE]] = []
        tables[column[ResponseKeys.TABLE]].append(column[ResponseKeys.COLUMN])

    unused_columns = []
    for table, columns in tables.items():
        table_unused_columns = [column[ResponseKeys.COLUMN] for column in unused_columns_and_tables if
                                column[ResponseKeys.TABLE] == table]
        if not table_unused_columns:
            continue
        columns.sort()
        table_unused_columns.sort()
        if table_unused_columns == columns:
            unused_columns.append(f'{table}: [.*]')
        elif any([False if column in table_unused_columns else True for column in columns]):
            unused_columns.append(f'{table}: [{", ".join(table_unused_columns)}]')

    for column in report_columns:
        column['used'] = False
        for unused_column in unused_columns_and_tables:
            if column[ResponseKeys.TABLE] == unused_column[ResponseKeys.TABLE] and column[ResponseKeys.COLUMN] == \
                    unused_column[ResponseKeys.COLUMN]:
                column[ResponseKeys.UNUSED] = True
                break

    return ', '.join(unused_columns), report_columns


def write_to_csv(file_path: str, values: List[List], overwrite: bool = False) -> None:
    """
    Writes values to a CSV file.

    Args:
        file_path (str): The path to the CSV file.
        values (List[str]): The values to write to the CSV file.
        overwrite (bool): Whether to overwrite the file or append to it. Default is False.
    """
    mode: str = 'w' if overwrite else 'a'
    with open(file_path, mode=mode, newline='', encoding='utf-8') as file:
        writer: csv.writer = csv.writer(file)
        writer.writerows(values)


def write_to_txt(file_path: str, values: List[str]) -> None:
    """
    Writes values to a text file.

    Args:
        file_path (str): The path to the text file.
        values (List[str]): The values to write to the text file.
    """
    with open(file_path, 'w', encoding='utf-8') as file:
        file.write('\n'.join(values))


def split_and_format(text):
    split_text = re.sub(r'(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])', ' ', text)
    return split_text



import argparse
import json
import os
from datetime import datetime, timedelta
from urllib.parse import urlparse

import msal
import requests
from colorama import Fore

from pb_analyzer.base_analyzer import BaseAnalyzer
from pb_analyzer.const import Requests, ResponseKeys, SHARED_TO_ORG_HEADERS
from pb_analyzer.utils import count_hidden_true_in_dict, fetch_columns_and_tables, get_classified_columns, \
    write_to_csv


class SharedReportsAnalyzer(BaseAnalyzer):
    def __init__(self, output_folder: str = None, debug: bool = False):
        """

        Args:
            output_folder: The path to the output folder.

        Example usage:
        SharedToWholeOrganizationAnalyzer('C:/Users/username/Downloads/Output.csv', 'C:/Users/username/Downloads/Results.txt').analyze()
        """
        time = int(round(datetime.now().timestamp()))
        is_default_result_path = True

        results_output_path = os.path.join(os.getcwd(), f'SharedReportsWithUnusedData_{time}.csv')
        summary_output_path = os.path.join(os.getcwd(), f'PBAnalyzerResults_{time}.txt')

        if output_folder:
            if '.' in os.path.basename(output_folder):
                print(Fore.RED + 'Invalid output folder path. Using default path.')
            if not os.path.isdir(output_folder):
                os.makedirs(output_folder)
            is_default_result_path = False
            results_output_path = os.path.join(output_folder, f'SharedReportsWithUnusedData_{time}.csv')
            summary_output_path = os.path.join(output_folder, f'PBAnalyzerResults_{time}.txt')

        super().__init__('Reports shared to whole organization analyzer', results_output_path, is_default_result_path,
                         summary_output_path, debug)

    def _extract_artifact_ids(self, response: dict):
        try:
            artifact_ids = [
                (entity[ResponseKeys.ARTIFACT_ID],
                 entity.get(ResponseKeys.SHARER, {}).get(ResponseKeys.DISPLAY_NAME, {}),
                 entity.get(ResponseKeys.DISPLAY_NAME, {})
                 ) for entity in
                response.get(ResponseKeys.ARTIFACT_ACCESS_ENTITIES, [])]
            if not artifact_ids:
                raise Exception('No reports shared to whole organization found.')
            return artifact_ids
        except Exception as e:
            print(Fore.RED + 'Failed to extract artifact IDs.')
            if self._debug:
                print(Fore.RED + str(e))
                print(Fore.RED + 'response:', response)
            exit(1)

    @staticmethod
    def _get_token():
        """
        Acquires an access token using the Microsoft Authentication Library (MSAL).
        """
        app = msal.PublicClientApplication(Requests.CLIENT_ID, authority=Requests.AUTHORITY)
        result = app.acquire_token_interactive(scopes=Requests.SCOPE, prompt='login')

        if ResponseKeys.ACCESS_TOKEN in result:
            return result[ResponseKeys.ACCESS_TOKEN]
        else:
            raise Exception("Failed to acquire token: %s" % result.get("error_description"))

    @staticmethod
    def _reports_published_to_web_api(token: str):
        headers = {'authorization': f'Bearer {token}'}
        response = requests.get(Requests.PUBLISHED_TO_WEB_URL, headers=headers)

        if response.status_code == 200:
            response_data = response.json()
            return response_data

    def _links_shared_to_whole_organization_api(self, token: str):
        headers = {'authorization': f'Bearer {token}'}
        response = requests.get(Requests.SHARED_TO_ORG_URL, headers=headers)

        if response.status_code == 200:
            response_data = response.json()
            if self._debug:
                print(Fore.GREEN + 'Successfully fetched shared reports.')
            return response_data
        else:
            raise Exception(f'Failed to get shared reports. Status code: {response.status_code}', response.json())

    @staticmethod
    def _send_push_access_request(token: str, report_id: str, region: str):
        headers = {'authorization': f'Bearer {token}'}
        response = requests.post(Requests.PUSH_ACCESS_URL.format(region, report_id), headers=headers)

        if response.status_code == 200:
            response_data = response.json()
            return response_data

    @staticmethod
    def _send_exploration_request(token: str, artifact_id: str, region: str):
        headers = {'authorization': f'Bearer {token}'}
        response = requests.get(Requests.EXPLORATION_URL.format(region, artifact_id), headers=headers)

        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f'Failed to send exploration request. Status code: {response.status_code}')

    @staticmethod
    def _send_conceptual_schema_request(token: str, model_id: str, region: str):
        headers = {'authorization': f'Bearer {token}', 'content-type': 'application/json; charset=UTF-8'}
        data = {"modelIds": [model_id], "userPreferredLocale": "en-US"}
        response = requests.post(Requests.CONCEPTUAL_SCHEMA_URL.format(region), headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f'Failed to send conceptual schema request. Status code: {response.status_code}')

    def _extract_region(self, response: dict):
        try:
            url = urlparse(response.get(ResponseKeys.REGION))
            region = url.netloc
            if self._debug:
                print(Fore.GREEN + f'Extracted region: {region}')
            return region
        except Exception as e:
            print(Fore.RED + 'Failed to extract region.')
            if self._debug:
                print(Fore.RED + str(e))
            exit(1)

    def _process_report(self, report, bar, start_time, *args):
        region, token = args
        report_id, sharer_name, name = report
        try:
            if datetime.now() - start_time > timedelta(minutes=10):
                print(Fore.RED + 'Passed the 10 minutes mark. Stopped the analysis.')
                raise TimeoutError('Passes the 10 minutes mark.')

            response_data = self._send_push_access_request(token, report_id, region)
            if response_data:
                conceptual_schema, exploration_response = self._get_report_data(region, response_data, token)
                num_of_hidden = count_hidden_true_in_dict(conceptual_schema)
                columns_and_tables = fetch_columns_and_tables(conceptual_schema)
                unused_message, all_columns = get_classified_columns(columns_and_tables, exploration_response)
                self._insert_all_columns(all_columns, num_of_hidden)
                self._success_count += 1
                if unused_message:
                    self._reports_with_unused_columns += 1
                    self._results.append([report_id, name, sharer_name, num_of_hidden, unused_message])
        except Exception as e:
            if self._debug:
                print(Fore.RED + str(e))
        bar()

    def _intro(self):
        self._handle_input()
        print(
            Fore.CYAN + 'This tool will analyze reports shared to the whole organization. You will need to sign in to '
                        'your Power BI account to proceed.')
        input(Fore.CYAN + 'Press Enter to start the analysis process..')
        token = self._get_token()
        all_shared_res = self._links_shared_to_whole_organization_api(token)
        reports = self._extract_artifact_ids(all_shared_res)
        print(Fore.GREEN + f'Found {len(reports)} reports shared to whole organization.')
        region = self._extract_region(all_shared_res)
        write_to_csv(self._result_output_path, [SHARED_TO_ORG_HEADERS], True)
        average_time_per_report = 0.5
        estimated_time = len(reports) * average_time_per_report
        print(Fore.YELLOW + f'Estimated time to analyze all reports: {estimated_time} seconds.')
        return region, reports, token

    def _get_report_data(self, region, response_data, token):
        artifact_id = response_data.get(ResponseKeys.ENTITY_KEY, {}).get(ResponseKeys.ID)
        model_id = next(item.get(ResponseKeys.ID) for item in response_data.get(ResponseKeys.RELATED_ENTITY_KEY) if
                        item.get(ResponseKeys.TYPE) == 4)
        conceptual_schema = self._send_conceptual_schema_request(token, model_id, region)
        exploration_response = self._send_exploration_request(token, artifact_id, region)
        return conceptual_schema, exploration_response

    def analyze(self):
        try:
            self._welcome_text()
            region, reports, token = self._intro()
            end_time, reports, start_time = (self._run_analysis(reports, region, token))
            self._outro(end_time, reports, start_time)
        except Exception as e:
            print(Fore.RED + 'Exiting due to an error.')
            if self._debug:
                print(Fore.RED + str(e))


def main():
    parser = argparse.ArgumentParser(description='Analyze Shared Power BI Reports')
    parser.add_argument('--output-folder', type=str, help='The path to the output folder')
    parser.add_argument('--debug', action='store_true', help='Enable debug mode')
    args = parser.parse_args()

    analyzer = SharedReportsAnalyzer(
        output_folder=args.output_folder,
        debug=args.debug
    )
    analyzer.analyze()



import argparse
import base64
import csv
import json
import os
import re
from datetime import datetime, timedelta
from typing import List, Optional, Dict, Union
from urllib.parse import urlparse, parse_qs, ParseResult

import requests
from colorama import Fore

from pb_analyzer.base_analyzer import BaseAnalyzer
from pb_analyzer.const import PublicRequests, REGEX_BI_REQUEST, NEW_HEADERS, ResponseKeys, ExplorationRequestError
from pb_analyzer.utils import count_hidden_true_in_dict, get_classified_columns, \
    write_to_csv, fetch_columns_and_tables


class PublicReportsAnalyzer(BaseAnalyzer):
    def __init__(self, embed_codes_file_path: str, output_folder: str = None, debug: bool = False):
        """
        Args:
            embed_codes_file_path:  The full path to the Power BI Reports CSV file.
            output_folder: The path to the output folder.

        Example usage:
        PublicReportAnalyzer('C:/Users/username/Downloads/PowerBIReports.csv', 'C:/Users/username/Downloads/Output.csv').analyze()
        """
        time = int(round(datetime.now().timestamp()))
        try:
            self._validate_embed_codes_file(embed_codes_file_path)
        except (FileNotFoundError, ValueError) as e:
            print(Fore.RED + str(e))
            exit(1)
        self._embed_codes_file_path = embed_codes_file_path
        is_default_results_path = True

        results_output_path = os.path.join(os.getcwd(), f'PublicReportsWithUnusedData_{time}.csv')
        summary_output_path = os.path.join(os.getcwd(), f'PBAnalyzerResults_{time}.txt')

        if output_folder:
            if '.' in os.path.basename(output_folder):
                print(Fore.RED + 'Invalid output folder path. Using default path.')
            if not os.path.isdir(output_folder):
                os.makedirs(output_folder)
            is_default_results_path = False
            results_output_path = os.path.join(output_folder, f'PublicReportsWithUnusedData_{time}.csv')
            summary_output_path = os.path.join(output_folder, f'PBAnalyzerResults_{time}.txt')

        super().__init__('Analyze Public Reports', results_output_path,
                         is_default_results_path, summary_output_path, debug)

    @staticmethod
    def _validate_embed_codes_file(embed_codes_path: str):
        if not os.path.isfile(embed_codes_path):
            raise FileNotFoundError(f'File not found: {embed_codes_path}')

        if not embed_codes_path.lower().endswith('.csv'):
            raise ValueError(f'File must be a CSV: {embed_codes_path}')

        expected_headers = ['Report name', 'Workspace name', 'Published by', 'Status', 'Embed URL']

        with open(embed_codes_path, mode='r', encoding='utf-8-sig') as file:
            reader = csv.reader(file)
            headers = next(reader, None)

            if headers != expected_headers:
                raise ValueError(
                    f'CSV file does not have the expected headers. Are you sure this is an Embed Codes CSV file?')

    @staticmethod
    def _send_power_bi_request(url: str) -> Optional[str]:
        response: requests.Response = requests.get(url)
        if response.status_code == 200:
            match: Optional[re.Match] = re.search(REGEX_BI_REQUEST, response.text)
            if match:
                return match.group(1)

    @staticmethod
    def _send_exploration_request(region_url: str, resource_key: str) -> Optional[Dict]:
        headers: Dict[str, str] = {'X-PowerBI-ResourceKey': resource_key}
        response: requests.Response = requests.get(
            PublicRequests.EXPLORATION_URL.format(region_url, resource_key),
            headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise ExplorationRequestError(f'Failed to get exploration data. Status code: {response.status_code}',
                                          response.json()['error']['code'])

    @staticmethod
    def _get_model_id(json_response: Union[str, Dict]) -> Optional[str]:
        try:
            data: Dict = json.loads(json_response) if isinstance(json_response, str) else json_response
            models: List[Dict] = data.get(ResponseKeys.MODELS, [{}])
            if models:
                return models[0].get(ResponseKeys.ID)
        except (json.JSONDecodeError, KeyError, TypeError) as e:
            return None

    @staticmethod
    def _send_conceptual_schema_request(region_url: str, model_id: str, resource_key: str) -> Optional[Dict]:
        headers: Dict[str, str] = {'X-PowerBI-ResourceKey': resource_key}
        payload: Dict[str, List[str]] = {'modelIds': [model_id]}
        response: requests.Response = requests.post(PublicRequests.CONCEPTUAL_SCHEMA_URL.format(region_url),
                                                    headers=headers, json=payload)
        if response.status_code == 200:
            return response.json()
        else:
            response.raise_for_status()

    @staticmethod
    def _decode_url(encoded_url: str) -> dict:
        parsed_url: ParseResult = urlparse(encoded_url)
        encoded_bytes: str = parse_qs(parsed_url.query)['r'][0]
        decoded_bytes: bytes = base64.b64decode(encoded_bytes)
        decoded_url: str = decoded_bytes.decode('utf-8')
        return json.loads(decoded_url)

    def _process_report(self, row, bar, start_time, *args):
        try:
            if datetime.now() - start_time > timedelta(minutes=10):
                raise TimeoutError('Passes the 10 minutes mark.')

            conceptual_schema, exploration_response = self._get_report_data(row)
            report_columns = fetch_columns_and_tables(conceptual_schema)
            num_of_hidden = count_hidden_true_in_dict(conceptual_schema)
            report_with_unused, all_columns = get_classified_columns(report_columns, exploration_response)
            self._insert_all_columns(all_columns, num_of_hidden)
            self._success_count += 1
            if report_with_unused:
                self._reports_with_unused_columns += 1
                self._results.append(row[:-1] + [num_of_hidden, report_with_unused])
        except ExplorationRequestError as e:
            if self._debug:
                print(Fore.RED + str(e))
            self._errors.append([row[0], e.args[1]])
        except TimeoutError as e:
            raise e
        except Exception as e:
            if self._debug:
                print(Fore.RED + str(e))
        bar()
        return

    def _intro(self):
        self._handle_input()
        rows = []
        with open(self._embed_codes_file_path, mode='r', encoding='utf-8') as file:
            reader = csv.reader(file)
            headers = next(reader)
            [rows.append(row) for row in reader]
        write_to_csv(self._result_output_path, [headers + NEW_HEADERS], True)
        average_time_per_report = 0.4
        estimated_time = round(len(rows) * average_time_per_report, 2)
        print(Fore.GREEN + f'Found {len(rows)} reports to analyze.')
        print(Fore.YELLOW + f'Estimated time to analyze all reports: {estimated_time} seconds.')
        if estimated_time > 60 * 10:
            print(Fore.RED + f'Scan will stop after approximately 10 minutes.')
        input(Fore.CYAN + "Press Enter to start the analysis...")
        print()
        print(Fore.BLUE + 'Analyzing reports...')
        return rows

    def _get_report_data(self, row):
        region_url = self._send_power_bi_request(row[4])
        region_url = region_url.replace('redirect', 'api')
        decoded_url = self._decode_url(row[4])
        exploration_response = self._send_exploration_request(region_url, decoded_url['k'])
        model_id = self._get_model_id(exploration_response)
        conceptual_schema = self._send_conceptual_schema_request(region_url, model_id, decoded_url['k'])
        return conceptual_schema, exploration_response

    def analyze(self):
        try:
            self._welcome_text()
            rows = self._intro()
            end_time, rows, start_time = self._run_analysis(rows)
            self._outro(end_time, rows, start_time)
        except Exception as e:
            print(Fore.RED + 'Exiting due to an error.')
            if self._debug:
                print(Fore.RED + str(e))


def main():
    parser = argparse.ArgumentParser(description='Analyze Public Power BI Reports')
    parser.add_argument('--embed-codes-path', type=str, help='Path to the Power BI Reports CSV file', required=True)
    parser.add_argument('--output-folder', type=str, help='The path to the output folder')
    parser.add_argument('--debug', action='store_true', help='Enable debug mode')
    args = parser.parse_args()

    analyzer = PublicReportsAnalyzer(
        embed_codes_file_path=args.embed_codes_path,
        output_folder=args.output_folder,
        debug=args.debug
    )
    analyzer.analyze()



import concurrent.futures
import os.path
from abc import abstractmethod
from datetime import datetime
from time import sleep

from alive_progress import alive_bar
from colorama import Fore, init

from pb_analyzer.const import ResponseKeys
from pb_analyzer.utils import write_to_txt, write_to_csv, split_and_format


class BaseAnalyzer:
    def __init__(self, tool: str, result_output_path: str, is_default_results_path: bool, results_output_path: str,
                 debug: bool = False):
        self._debug = debug
        self._result_output_path = result_output_path
        self._summary_output_path = results_output_path
        self._is_default_results_path = is_default_results_path
        self._results = []
        self._tool = tool
        self._unused_columns = []
        self._total_columns_scanned = 0
        self._reports_with_hidden_columns = 0
        self._reports_with_unused_columns = 0
        self._success_count = 0
        self._errors = []

    def _insert_all_columns(self, all_columns, num_of_hidden):
        if num_of_hidden > 0:
            self._reports_with_hidden_columns += 1

        for column in all_columns:
            if not any(existing_column[ResponseKeys.TABLE] == column[ResponseKeys.TABLE] and existing_column[
                ResponseKeys.COLUMN] == column[ResponseKeys.COLUMN] for
                       existing_column in self._unused_columns):
                self._unused_columns.append(column)
            else:
                for existing_column in self._unused_columns:
                    if (existing_column[ResponseKeys.TABLE] == column[ResponseKeys.TABLE]
                            and existing_column[ResponseKeys.COLUMN] == column[ResponseKeys.COLUMN]):
                        existing_column[ResponseKeys.UNUSED] = True

    @staticmethod
    def _insert_new_path(path, suffix: str):
        print(Fore.YELLOW + f'Default output {suffix.upper()} file: {path}')
        print(Fore.YELLOW + f'Press Enter to continue with the default path or type a new path.')
        user_input = input()
        if user_input:
            if user_input.endswith(f'.{suffix}'):
                print(Fore.YELLOW + f'Output {suffix.upper()} file: {user_input}')
                return user_input
            else:
                print(Fore.RED + 'Invalid file path. Using default path.')
        return path

    @abstractmethod
    def _intro(self):
        pass

    def _handle_input(self):
        if not self._is_default_results_path:
            return

        print(Fore.YELLOW + f'Default output directory: {os.path.dirname(self._result_output_path)}')
        print(Fore.YELLOW + f'Press Enter to continue with the default directory or type a new directory.')
        user_input = input()
        if user_input:
            if '.' in user_input.split('/')[-1]:
                print(Fore.RED + 'Invalid directory. Using default paths.')
                return
            if os.path.isdir(user_input) or not os.path.exists(user_input):
                os.makedirs(user_input, exist_ok=True)
                self._result_output_path = os.path.join(user_input, os.path.basename(self._result_output_path))
                self._summary_output_path = os.path.join(user_input, os.path.basename(self._summary_output_path))
        print(Fore.GREEN + f'Output directory: {os.path.dirname(self._result_output_path)}')
        print()

    @abstractmethod
    def _process_report(self, row, bar, start_time, *args):
        raise NotImplementedError

    def _run_analysis(self, rows, *args):
        start_time = datetime.now()
        timed_out = False

        with alive_bar(len(rows), bar='blocks') as bar:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = [executor.submit(self._process_report, row, bar, start_time, *args) for row in rows]
                for future in concurrent.futures.as_completed(futures):
                    try:
                        future.result()
                    except TimeoutError as e:
                        if self._debug:
                            print(Fore.RED + str(e))
                        timed_out = True
                        executor.shutdown(wait=False, cancel_futures=True)
                        break
            write_to_csv(self._result_output_path, self._results)
        if timed_out:
            print(Fore.RED + 'Passed the 10 minutes mark. Stopped the analysis.')
            sleep(1)
        end_time = datetime.now()
        return end_time, rows, start_time

    def _welcome_text(self):
        init(autoreset=True)

        print(Fore.CYAN + "=" * 65)
        print(Fore.YELLOW + "Welcome to Power BI Analyzer - Report Analysis Tool".center(65))
        print(Fore.CYAN + "=" * 65)
        sleep(0.3)
        print()
        print(Fore.GREEN + "Project: Power BI Analyzer")
        sleep(0.3)
        print(Fore.GREEN + f"Tool: {self._tool}")
        print()
        sleep(0.3)
        print(Fore.WHITE + "This tool is part of the Power BI Analyzer project, which aims to help")
        sleep(0.2)
        print(Fore.WHITE + "organizations identify unused data sources in their Power BI reports.")
        sleep(0.2)
        print(Fore.WHITE + "Unused columns in your reports can pose a security risk, and it is")
        sleep(0.2)
        print(Fore.WHITE + "essential to identify and remove them to prevent data breaches.")
        print()
        sleep(0.5)
        print(Fore.MAGENTA + "BACKGROUND:")
        sleep(0.1)
        print(Fore.WHITE + "On June 19, 2024, Nokod Security published a warning about a data leakage")
        sleep(0.1)
        print(Fore.WHITE + "vulnerability in the Microsoft Power BI service. For more details, visit:")
        sleep(0.1)
        print(
            Fore.BLUE + "https://nokodsecurity.com/blog/in-plain-sight-how-microsoft-power-bi-reports-expose-sensitive-data-on-the-web/")
        print()
        sleep(0.5)
        print(Fore.WHITE + "Nokod Security created the \"Power BI Analyzer\" as a simple and free tool")
        sleep(0.1)
        print(Fore.WHITE + "for organizations to assess their exposure to this vulnerability. If you")
        sleep(0.1)
        print(Fore.WHITE + "need help with this tool, please contact amichai@nokodsecurity.com or")
        sleep(0.1)
        print(Fore.WHITE + "uriya@nokodsecurity.com.")
        print(Fore.CYAN + "=" * 65)
        sleep(0.5)

    def _outro(self, end_time, rows, start_time):
        print()
        print(Fore.CYAN + "=" * 65)
        print(Fore.MAGENTA + "Results".center(65))
        print(Fore.CYAN + "=" * 65)
        results = [
            f'Number of reports analyzed successfully: {self._success_count}/{len(rows)}',
            'Total tables scanned: ' + str(len(set([column['table'] for column in self._unused_columns]))),
            'Unique columns scanned: ' + str(len(self._unused_columns)),
            'Unused columns found: ' + str(
                len([column for column in self._unused_columns if column.get(ResponseKeys.UNUSED)])),
            'Reports with unused columns: ' + str(self._reports_with_unused_columns),
            'Reports with hidden columns: ' + str(self._reports_with_hidden_columns),
            'Scan time: ' + str(end_time - start_time),
            ''
        ]

        for line in results:
            print(Fore.WHITE + line)

        error_messages = []
        if self._success_count != len(rows):
            if self._errors:
                error_messages = [f'Failed to analyze "{row[0]}". Error: "{split_and_format(row[1])}"' for row in
                                  self._errors]

        title = ['', 'Project: Power BI Analyzer', f'Tool: {self._tool}', '']
        header = ['=' * 65, 'Results'.center(65), '=' * 65, '']
        closing = ['=' * 65]
        footer = ['', 'Full analysis saved to ' + self._result_output_path]
        write_to_txt(self._summary_output_path, title + header + results + closing + error_messages + footer)
        print(Fore.CYAN + "=" * 65)
        print()
        [print(Fore.RED + line) for line in error_messages]
        print()
        print(Fore.GREEN + 'Full analysis saved to ' + self._result_output_path)
        print(Fore.GREEN + 'Results saved to ' + self._summary_output_path)



SHARED_TO_ORG_HEADERS = ['Report Id', 'Report Name', 'Shared by', 'Number of hidden columns', 'Unused columns']
NEW_HEADERS = ['Number of hidden columns', 'Unused columns']
REGEX_BI_REQUEST = r"var resolvedClusterUri = 'https://(.*?)';"


class Requests:
    AUTHORITY = "https://login.microsoftonline.com/common"
    CLIENT_ID = "23d8f6bd-1eb0-4cc2-a08c-7bf525c67bcd"
    SCOPE = ['https://analysis.windows.net/powerbi/api/.default openid profile offline_access']
    PUBLISHED_TO_WEB_URL = 'https://api.powerbi.com/v1.0/myorg/admin/widelySharedArtifacts/publishedToWeb'
    SHARED_TO_ORG_URL = 'https://api.powerbi.com/v1.0/myorg/admin/widelySharedArtifacts/linksSharedToWholeOrganization'
    PUSH_ACCESS_URL = 'https://{}/metadata/access/reports/{}/pushaccess?forceRefreshGroups=true'
    EXPLORATION_URL = 'https://{}/explore/reports/{}/exploration'
    CONCEPTUAL_SCHEMA_URL = 'https://{}/explore/conceptualschema'


class PublicRequests(Requests):
    EXPLORATION_URL = 'https://{}/public/reports/{}/modelsAndExploration?preferReadOnlySession=true'
    CONCEPTUAL_SCHEMA_URL = 'https://{}/public/reports/conceptualschema'


class ResponseKeys:
    ACCESS_TOKEN = "access_token"
    ARTIFACT_ID = 'artifactId'
    ARTIFACT_ACCESS_ENTITIES = 'ArtifactAccessEntities'
    REGION = '@odata.context'
    REPORT = 'report'
    OWNER_INFO = 'ownerInfo'
    FIRST_NAME = 'givenName'
    LAST_NAME = 'familyName'
    DISPLAY_NAME = 'displayName'
    SHARER = 'sharer'
    UNKNOWN_REPORT = 'Unknown Report'
    UNKNOWN_SHARER = 'Unknown Sharer'
    TABLE = 'table'
    COLUMN = 'column'
    UNUSED = 'unused'
    MODELS = 'models'
    ID = 'id'
    ENTITY_KEY = 'entityKey'
    RELATED_ENTITY_KEY = 'relatedEntityKeys'
    TYPE = 'type'


class ExplorationRequestError(Exception):
    pass



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='power_bi_analyzer',
    version='1.0.3',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'public-reports-analyzer=pb_analyzer.power_bi_public_report_analyzer:main',
            'shared-reports-analyzer=pb_analyzer.power_bi_shared_reports_analyzer:main',
        ]
    },
    author='Nokod Security',
    author_email='support@nokodsecurity.com',
    description='A package to analyze Power BI reports.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Nokod/PBAnalyzer',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=required,
)
